const Alert = require('bem-alert');
const co = require('co');
const thunkify = require('thunkify');
const request = thunkify(require('request'));
const logger = require('log4js').getLogger('sales:jd');
const iconv = require('iconv-lite');
const URL = require('url');
const promCalc = require('prom-calc');
const redis = require('bem-redis');
const PRE = 'hx-crawl-jd';
const catMap = require('./catMap');
const moment = require('moment');
const promAnalysis = require('prom-analysis');
const md5 = require('bem-md5');
const async = require('async');

let Bem = null;

class CrawlJd {

  constructor(opt) {
    opt = opt || {};
    opt.redis = opt.redis || {};
    this._client = redis.get();
    Bem = opt.Bem;
  }

  * alert(msg) {
    return;
    if (!this._alert) {
      this._alert = new Alert({
        bem: {
          appId: 'dmVkYGPFKwMuCEZsU2mwEmpTkprNVQdM',
          apiKey: 'icUcv7SbfaZbBF5QkggCuTzzaqzs7uSq'
        },
        weixin: {
          appId: 'wx88d545a5def0d88f'
        }
      });
    }
    yield this._alert.send(msg);
  }

  get MAX_TRY() {
    return 3;
  }

  get blacklist() {
    return [
      'zbbs.jd.com',
      'home.jd.com',
      'vip.jd.com',
      'app.jd.com',
      'c-nfa.jd.com',
      'cart.jd.com',
      'b.jd.com',
      'gate.jd.com',
      'order.jd.com',
      'mall.jd.com',
      'help.jd.com',
      'chat.jd.com',
      'fw.jd.com',
      'boss.jd.com',
      'zhaopin.jd.com',
      'bang.jd.com',
      'zx.jd.com',
      'group.jd.com',
      'chat15.jd.com',
      'paimai.jd.com',
      'fuwu.jd.com',
      'giftcard.jd.com',
      'gongyi.jd.com',
      'jingzhi.jd.com',
      'ur.jr.jd.com',
      'm.jr.jd.com',
      'jr.jd.com',
      'jzt.jd.com',
      'licai.jd.com',
      'media.jd.com',
      'sale.jd.com',
      'jd2008.jd.com',
      'club.jd.com',
      'club.jr.jd.com',
      'coin.jd.com',
      'credit.jd.com',
      'cread.jd.com',
      'dang.jd.com',
      'jdd.jr.jd.com',
      'jdread.jd.com',
      'jimi.jd.com',
      'jiayouka.jd.com',
      'jipiao.jd.com',
      'jincai.jd.com',
      'jmall.jd.com',
      'jrb2b.jd.com',
      'jrhelp.jd.com',
      'jshop2013.jd.com',
      'licaishi.jd.com',
      'loan.jd.com',
      'menpiao.jd.com',
      'mobile.jd.com',
      'movie.jd.com',
      'myjd.jd.com',
      'm.z.jd.com',
      'nj.jd.com',
      'o.jd.com',
      'piao.jd.com',
      'q.jd.com',
      'qslc.jd.com',
      'quant.jd.com',
      'rich.jd.com',
      'saishi.jd.com',
      'sa.jd.com',
      'sfy.jd.com',
      'shipingou.jd.com',
      'survadmin.jd.com',
      'surveys.jd.com',
      'train.jd.com',
      'trip.jd.com',
      's.jd.com',
      'up.jd.com',
      'vip.jr.jd.com',
      'visa.jd.com',
      'wan.jd.com',
      'xiaobai.jr.jd.com',
      'xiaoyuan.jd.com',
      'xihuan.jd.com',
      'xjk.jr.jd.com',
      'xuanhao.jd.com',
      'y.jd.com',
      'you.jd.com',
      'youlun.jd.com',
      'zan.jd.com',
      'z.jd.com',
      'my.jd.com',
      'list.jr.jd.com',
      'dapeigou.jd.com',
      'dcrz.jd.com',
      'e.jd.com',
      'fang.jd.com',
      'game.jd.com',
      'fund.jd.com',
      'ft.jd.com',
      'gb.jd.com',
      'gold.jr.jd.com',
      'gupiao.jd.com',
      'hanyuan.jd.com',
      'hotel.jd.com',
      'huan.jd.com',
      'ihome.jd.com',
      'i.jd.com',
      't.jd.com',
      '8.jd.com',
      '95005.jd.com',
      'bbs.jd.com',
      'bill.jr.jd.com',
      'bk.jd.com',
      'caipiao.jd.com',
      'c.fa.jd.com',
      'card.jd.com',
      'chat5.jd.com',
      'cread.e.jd.com',
      'search-e.jd.com',
      'trade-e.jd.com',
      'chedai.jd.com',
      'chongzhi.jd.com',
      'en.jd.com'
    ];
  }

  * getPageContent(url, options, tried) {
    let charset
    if (typeof options === 'string') {
      charset = options
    } else {
      options = options || {}
      charset = options.charset
    }
    let defaultCharset = charset || 'gbk';
    tried = tried || 0;
    if (tried > this.MAX_TRY) {
      return null;
    }
    let result;
    try {
      result = yield request({
        encoding: null,
        gzip: true,
        timeout: 3000,
        headers: Object.assign({
          'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.98 Safari/537.36',
          'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8',
          'Accept-Language': 'en-US,en;q=0.8,zh-CN;q=0.6,zh;q=0.4,zh-TW;q=0.2,ja;q=0.2',
          'Cache-Control': 'max-age=0',
          'Upgrade-Insecure-Requests': '1',
          'Cookie': ''
        }, options.headers || {}),
        url
      });
    } catch (e) {
      console.log(url, e);
      return yield this.getPageContent(url, options, tried + 1);
    }
    let contentType = result[0].headers['content-type'];
    if (contentType) {
      charset = contentType.replace(/^.*charset=/, '').toLowerCase();
    }
    let body = null;
    switch (charset) {
      case 'utf-8':
      case 'gbk':
        break;
      default:
        let part = result[1].toString().substring(0, 200);
        if (part.indexOf('charset=utf-8') !== -1) {
          charset = 'utf-8'
        } else {
          charset = defaultCharset;
        }
        break;
    }
    body = iconv.decode(result[1], charset);
    return body;
  }

  * cache(key, value, time) {
    if (arguments.length === 1) {
      if (process.env.HX_TYPE === 'sub') {
        return null;
      }
      let cached = yield thunkify(this._client.get).call(this._client, `${PRE}-${key}`);
      return cached ? JSON.parse(cached) : null;
    } else {
      yield thunkify(this._client.set).call(this._client, `${PRE}-${key}`, JSON.stringify(value));
      yield thunkify(this._client.expire).call(this._client, `${PRE}-${key}`, time || 10);
    }
  }

  * getItemPriceFromRe(itemId, tried) {
    tried = tried || 0;
    if (tried > this.MAX_TRY) {
      return null;
    }
    let url = `http://re.m.jd.com/cps/item/${itemId}.html`;
    let result;
    try {
      result = yield request({
        timeout: 3000,
        gzip: true,
        jar: true,
        headers: {
          'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/57.0.2987.133 Safari/537.36'
        },
        url
      });
      result = result[1];
      let pattern = /var data = ([\s\S]*?) \<\/script\>/;
      let matched = result.match(pattern);
      if (matched) {
        let script = matched[1];
        let data = JSON.parse(script);
        return parseFloat(data.item.price);
      } else {
        return yield this.getItemPriceFromRe(itemId, tried + 1);
      }
    } catch (e) {
      return yield this.getItemPriceFromRe(itemId, tried + 1);
    }
  }

  * getItemPriceFromPage(itemId, tried) {
    tried = tried || 0;
    if (tried > this.MAX_TRY) {
      return null;
    }
    let cached = yield this.cache(`price-${itemId}`);
    if (cached) {
      return cached;
    }
    let price = yield this.getItemPriceFromRe(itemId);
    if (price) {
      yield this.cache(`price-${itemId}`, price);
      return price;
    }
    let url = `http://item.m.jd.com/product/${itemId}.html`
    let result;
    try {
      result = yield request({
        timeout: 3000,
        gzip: true,
        jar: true,
        headers: {
          'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/57.0.2987.133 Safari/537.36'
        },
        url
      });
      result = result[1];
      let pattern = /<input type="hidden" id="jdPrice" name="jdPrice" value="([\d\.]+)"\/>/
      let matched = result.match(pattern);
      if (matched) {
        yield this.cache(`price-${itemId}`, matched[1]);
        return parseFloat(matched[1]);
      } else {
        return yield this.getItemPriceFromPage(itemId, tried + 1);
      }
    } catch (e) {
      logger.error(e);
      return yield this.getItemPrice(itemId, tried + 1);
    }
  }

  * setOffline(item) {
    if (!item.get('offline')) {
      item.set('offline', true);
      yield item.save();
    }
  }

  * getItemPrice(item, tried) {
    if (item.get('offline')) {
      return null;
    }
    let itemId = item.get('jdId');
    let cached = yield this.cache(`price-${itemId}`);
    if (cached) {
      return cached;
    }
    tried = tried || 0;
    if (tried > this.MAX_TRY) {
      return null;
    }
    let price = yield this.getItemPriceFromRe(itemId);
    if (price) {
      yield this.cache(`price-${itemId}`, price);
      return price;
    }
    // let url = `http://p.3.cn/prices/gets?pduid=${parseInt(Math.random()*100000000)}&skuid=J_${itemId}`;
    let url = `http://item.m.jd.com/product/${itemId}.html`
    let result;
    try {
      result = yield request({
        timeout: 3000,
        gzip: true,
        jar: true,
        url
      });
      result = result[1];
      let pattern = /<input type="hidden" id="jdPrice" name="jdPrice" value="([\d\.]+)"\/>/
      let matched = result.match(pattern);
      if (matched) {
        yield this.cache(`price-${itemId}`, matched[1]);
        return parseFloat(matched[1]);
      } else {
        let isOffline = this.checkOffline(result);
        if (isOffline) {
          yield this.setOffline(item);
          return null;
        } else {
          return yield this.getItemPrice(item, tried + 1);
        }
      }
    } catch (e) {
      logger.error(e);
      return yield this.getItemPrice(item, tried + 1);
    }
  }

  calcLastPrice(item, historyPrices) {
    if (!historyPrices) return;
    let lastPrice = item.get('lastPrice');
    let yesterday = moment().subtract(1, 'days').format('YYYY-MM-DD');
    let yesterdayHistory = historyPrices[yesterday];
    if (yesterdayHistory) {
      if (yesterdayHistory.get('price') !== lastPrice) {
        lastPrice = yesterdayHistory.get('price');
        item.set('lastPrice', parseFloat(lastPrice));
      }
    }
  }

  * getHistoryPrices(items) {
    let query = new Bem.Query('PriceHistory');
    let startTime = moment().subtract(6, 'months').toDate();
    let history = yield query.containedIn('item', items).include('item').greaterThanOrEqualTo('createdAt', startTime).find();
    let histories = {};
    history.forEach(one => {
      let item = one.get('item');
      let id = item.id;
      if (!histories[id]) histories[id] = {};
      let date = moment(one.get('createdAt')).format('YYYY-MM-DD');
      histories[id][date] = one;
    });
    return histories;
  }

  * getSubscriptions(items) {
    let query = new Bem.Query('Subscription');
    let subs = yield query.containedIn('item', items).find({
      useMasterKey: true
    });
    let ret = {};
    subs.forEach(sub => {
      let id = sub.get('item').id;
      if (!ret[id]) ret[id] = [];
      ret[id].push(sub);
    });
    return ret;
  }

  * getSubscriptionLastFeed(sub) {
    let user = sub.get('user');
    let uid = user.id;
    let item = sub.get('item');
    let id = item.id;
    let feedsQuery = new Bem.Query('SubscriptionFeed');
    let start = new Date(`${moment().format('YYYY-MM-DD')} 00:00:00`);
    let lastFeed = yield feedsQuery.equalTo('item', item).equalTo('user', user).greaterThanOrEqualTo('createdAt', start).find();
    return lastFeed;
  }

  * setSubscriptionFeedsExpired(items, subs, feeds) {
    let feedsExpired = [];
    for (let item of items) {
      let itemSubs = subs;
      for (let sub of itemSubs) {
        let user = sub.get('user');
        let feedsQuery = new Bem.Query('SubscriptionFeed');
        let oldFeeds = yield feedsQuery.equalTo('item', item).equalTo('user', user).equalTo('expired', null).notContainedIn('objectId', feeds.map(feed => feed.id)).find();
        if (oldFeeds && oldFeeds.length > 0) {
          for (let feed of oldFeeds) {
            let oldProms = feed.get('proms');
            feed.set('expired', true);
            feedsExpired.push(feed);
          }
        }
      }
    }
    yield Bem.Object.saveAll(feedsExpired);
  }

  isSameProms(prom1, prom2) {
    prom1 = prom1 || [];
    prom2 = prom2 || [];
    //同为null或undefined
    if (prom1 === prom2) return true;
    else if (prom1.length === prom2.length) {
      return prom1.every((prom, i) => {
        return prom2.some(p => {
          if (prom.name === '领券') {
            return prom.desc === p.desc;
          } else if (prom.name === '降价') {
            return p.name === '降价' && prom.price === p.price;
          } else {
            return prom.desc === p.desc;
          }
        })
      });
    }
    return false;
  }

  getMinPrice(item) {
    let price = parseFloat(item.get('price'));
    let proms = item.get('proms');
    if (!proms || proms.length === 0) {
      return price;
    }
    let minPrice = price;
    proms.forEach(prom => {
      let promPrice = parseFloat(prom.price);
      if (promPrice && promPrice < minPrice) {
        minPrice = promPrice
      }
    });
    return minPrice;
  }

  * createFeed(item, proms, promChanged, subs, historyPrices) {
    historyPrices = historyPrices || {};
    let ret = [];
    for (let sub of subs) {
      let isNewSub = sub.get('sent') !== true;
      if (isNewSub || promChanged) {
        sub.set('sent', true);
        let user = sub.get('user');
        let lastFeed = yield this.getSubscriptionLastFeed(sub);
        let isPushed = false;
        if (lastFeed && lastFeed.length > 0) {
          isPushed = lastFeed.some(feed => {
            return this.isSameProms(feed.get('proms'), proms);
          });
        }
        let yesterday = moment().subtract(1, 'days').format('YYYY-MM-DD');
        let lastHistory = historyPrices[yesterday];
        if (isPushed || (!isNewSub && lastHistory && lastHistory.get('minPrice') == this.getMinPrice(item))) {
          continue;
        }
        if (proms && proms.length > 0) {
          logger.info(`${item.get('name')}: 发送优惠通知`)
          let Feed = Bem.Object.extend('SubscriptionFeed');
          let feed = new Feed();
          feed.set('user', sub.get('user'));
          feed.set('item', item);
          feed.set('corp', sub.get('corp'));
          feed.set('proms', item.get('proms'));
          ret.push(feed);
        }
      }
    }
    yield Bem.Object.saveAll(ret, {
      useMasterKey: true
    });
    yield Bem.Object.saveAll(subs, {
      useMasterKey: true
    });
    return ret;
  }

  * savePriceHistory(item, historyPrices) {
    historyPrices = historyPrices || {};
    let price = item.get('price');
    let proms = item.get('proms');
    let minPrice = this.getMinPrice(item);
    let date = moment().format('YYYY-MM-DD');
    let todayHistory = historyPrices[date];
    let oldProms = [];
    let diff = 1;
    let yesterday = moment().subtract(diff, 'days').format('YYYY-MM-DD');
    let yesterdayHistory = historyPrices[yesterday];
    while (!yesterdayHistory && diff <= 180) {
      diff++;
      yesterday = moment().subtract(diff, 'days').format('YYYY-MM-DD');
      yesterdayHistory = historyPrices[yesterday];
    }
    if (!todayHistory && yesterdayHistory && this.isSameProms(yesterdayHistory.get('proms'), proms)) {
      return null;
    }
    if (todayHistory) {
      if (todayHistory.get('oldProms')) {
        oldProms = todayHistory.get('oldProms');
      }
      oldProms.push(todayHistory.get('proms'));
      todayHistory.set('oldProms', oldProms);
      todayHistory.set('proms', proms);
      todayHistory.set('name', item.get('name'));
      todayHistory.set('price', parseFloat(price));
      todayHistory.set('minPrice', parseFloat(minPrice));
    } else {
      let PriceHistory = Bem.Object.extend('PriceHistory');
      todayHistory = new PriceHistory();
      todayHistory.set('name', item.get('name'));
      todayHistory.set('item', item);
      todayHistory.set('price', parseFloat(price));
      todayHistory.set('minPrice', parseFloat(minPrice));
      todayHistory.set('proms', proms);
    }
    return todayHistory;
  }

  calcDiscount(item) {
    let proms = item.get('proms');
    let price = parseFloat(item.get('price'));
    let lastPrice = item.get('lastPrice');
    let oriPrice = (lastPrice && lastPrice > price) ? lastPrice : price;
    let minPrice = this.getMinPrice(item);
    let discount = parseFloat((minPrice / oriPrice * 10).toFixed(1));
    if (discount !== item.get('discount')) {
      item.set('discount', discount);
    }
  }

  calcPromScore(item, historyPrices, shopBlacklist) {
    historyPrices = historyPrices || {};
    let dates = Object.keys(historyPrices);
    if (dates.length === 0) return;
    dates.sort();
    let history = dates.map(date => historyPrices[date].toJSON());
    let result = promAnalysis(item.get('proms'), history);
    if (result.score !== item.get('promScore')) {
      item.set('promScore', result.score);
    }
    let isBlackShop = shopBlacklist.indexOf(item.get('jdShop')) !== -1;
    if (result.score >= 300 && !isBlackShop) {
      item.set('isNewLow', result.isNewLow)
      if (!item.get('isRecommend') === true) {
        item.set('isRecommend', true);
        item.set('recommendedAt', new Date().toISOString());
      }
    } else {
      item.set('isNewLow', false)
      if (!item.get('isRecommend') === false) {
        item.set('isRecommend', false);
      }
    }
  }

  * getShopBlacklist() {
    let shopQuery = new Bem.Query('ShopBlacklist');
    let blackShops = yield shopQuery.find();
    return blackShops.map(shop => shop.get('id'));
  }

  * saveItems(items) {
    yield Bem.Object.saveAll(items);
    for (let item of items) {
      let result = yield this.updateESIndex(item);
      if (!result || result.status === 400) {
        logger.error(item.id, result);
      }
    }
  }

  * ensureHistory(item, historyPrices) {
    let dates = Object.keys(historyPrices);
    if (dates.length === 0) return null;
    dates.sort();
    let lastHistory = historyPrices[dates[dates.length - 1]];
    if (!lastHistory || parseFloat(lastHistory.get('price')) !== parseFloat(item.get('price')) || parseFloat(lastHistory.get('minPrice')) !== parseFloat(this.getMinPrice(item))) {
      try {
        logger.info(`新增遗漏的价格记录: ${item.get('name')}(${item.get('jdId')})`);
        return yield this.savePriceHistory(item, historyPrices);
      } catch (e) {
        logger.error('save price failed:', e);
        return null;
      }
    }
  }

  * updateESIndex(item) {
    let data = item.toJSON();
    data.events = item.get('events') && item.get('events').map(ev => ev.toJSON())
    data.tags = item.get('tags') && item.get('tags').map(tag => tag.toJSON())
    data.cat = item.get('cat').map(cat => {
      let data = cat.toJSON()
      data.className = 'Categories'
      data.events = cat.get('events') && cat.get('events').map(ev => ev.toJSON())
      data.tags = cat.get('tags') && cat.get('tags').map(tag => tag.toJSON())
      return data
    })
    delete data.yesterdayHistory
    delete data.proms;
    delete data.lastProms;
    if (data.commision) {
      delete data.commision.endDate;
      delete data.commision.startDate;
      delete data.commision.updatedAt;
    }
    return (yield request({
      url: `http://${process.env.SEARCH_HOST||'192.168.7.8:9200'}/test_dmvkygpfkwmucezsu2mwemptkprnvqdm/Items/${item.id}`,
      method: 'PUT',
      json: data
    }))[1];
  }

  checkOffline(body) {
    return body.indexOf('该商品已下柜') !== -1;
  }

  * refreshItem(shopBlacklist, historyToSave, historyPrices, subs, item, context) {
    context.t_start = Date.now()
    // console.time(`fetch: ${item.id}`);
    let originItemData = item.toJSON();
    context.t_item_tojson = Date.now()
    // console.timeEnd(`fetch: ${item.id}`);
    // console.time(`get item data: ${item.id}`);
    let itemData = yield this.getItemData(item)
    context.t_get_item_data = Date.now()
    if (!itemData) {
      logger.error('crawl item data failed');
      item.set('offline', true);
      yield item.save();
      return;
    } else if (item.get('offline') === true) {
      item.set('offline', false);
    }
    // console.timeEnd(`get item data: ${item.id}`);
    if (itemData.name !== item.get('name')) {
      item.set('name', itemData.name);
    }
    let pic = itemData.pic.replace('120x120', '360x360');
    if (pic !== item.get('pic')) {
      item.set('pic', pic);
    }
    if (!item.get('jdCat') || item.get('jdCat').some(cat => {
        return cat === null;
      })) {
      item.set('jdCat', itemData.cat);
    }
    if (item.get('jdShop') !== itemData.shop) {
      item.set('jdShop', itemData.shop);
    }
    if (item.get('jdSelf') !== itemData.self) {
      item.set('jdSelf', itemData.self);
    }
    if (item.get('jdOP') !== itemData.op) {
      item.set('jdOP', itemData.op);
    }
    if (!item.get('brand')) {
      item.set('brand', itemData.brand);
    }
    if (!item.get('priceChangedAt')) {
      item.set('priceChangedAt', new Date().toISOString())
    }
    let Cat = Bem.Object.extend('Categories');
    let mappedCat = itemData.cat.map((cat) => {
      if (!catMap[cat]) {
        return;
      }
      let newCat = new Cat();
      newCat.id = catMap[cat];
      return newCat;
    }).filter((cat) => {
      return !!cat
    });
    if (!item.get('cat') || item.get('cat').length < mappedCat.length) {
      item.set('cat', mappedCat);
    }
    context.t_update_item_data = Date.now()
    // console.time(`getProm: ${item.id}`);
    let proms = yield this.getProms(item);
    context.t_get_prom = Date.now()
    // console.timeEnd(`getProm: ${item.id}`);
    // console.time(`getItemPrice: ${item.id}`);
    let price = yield this.getItemPrice(item);
    context.t_get_price = Date.now()
    //没取到价格应该是当前无货了，暂时仍按原价处理
    if (!price) {
      if (item.get('soldOut') !== true) {
        item.set('soldOut', true);
      }
      price = item.get('price');
    } else {
      if (item.get('soldOut') !== false) {
        item.set('soldOut', false);
      }
    }
    price = parseFloat(price);
    // console.timeEnd(`getItemPrice: ${item.id}`);
    let priceChanged = price !== item.get('price');
    this.calcLastPrice(item, historyPrices[item.id]);
    if (priceChanged) {
      item.set('price', price);
    }
    if (price < (item.get('lastPrice') || 0)) {
      proms = proms || [];
      proms.push({
        name: '降价',
        desc: `由${item.get('lastPrice')}元降为${price}元`,
        detail: {
          originPrice: item.get('lastPrice'),
          newPrice: price
        },
        t: (new Date()).toISOString(),
        price
      })
    }
    let promChanged = !this.isSameProms(item.get('proms'), proms);
    context.t_calc_prom = Date.now()
    //价格有变化，或者促销有变化时，记录一次价格历史
    logger.info(`${item.get('name')}: ${(proms && proms.length > 0) ? proms.map(prom=>prom.name) : '无促销'}(${promChanged?'有变化':'无变化'})`)
    if (priceChanged || promChanged) {
      item.set('priceChangedAt', new Date().toISOString())
      if (promChanged) {
        item.set('lastProms', item.get('proms'));
        item.set('proms', proms);
      } else if (priceChanged) {
        item.set('proms', proms);
      }
      if (process.env.HX_TYPE === 'sub' && subs[item.id]) {
        let createdFeeds = yield this.createFeed(item, proms, promChanged, subs[item.id], historyPrices[item.id]);
        feeds = feeds.concat(createdFeeds);
        if (createdFeeds.length > 0) {
          yield this.setSubscriptionFeedsExpired([item], subs[item.id], createdFeeds);
        }
      }
      let history = yield this.savePriceHistory(item, historyPrices[item.id]);
      if (history) {
        historyToSave.push(history);
        logger.info(`新增价格记录: ${item.get('name')}(${item.get('jdId')})`);
      }
    } else {
      if (process.env.HX_TYPE === 'sub' && subs[item.id]) {
        let createdFeeds = yield this.createFeed(item, proms, promChanged, subs[item.id], historyPrices[item.id]);
        feeds = feeds.concat(createdFeeds);
      }
    }
    context.t_create_history = Date.now()
    this.calcDiscount(item, proms);
    this.calcPromScore(item, historyPrices[item.id], shopBlacklist);
    context.t_calc_prom_score = Date.now()
  }

  parallelRefreshItems(items, options) {
    let shopBlacklist = options.shopBlacklist;
    let historyToSave = options.historyToSave;
    let historyPrices = options.historyPrices;
    let itemsToRetry = options.itemsToRetry;
    let subs = options.subs;
    return (done) => {
      async.each(items, (item, callback) => {
        try {
          let _this = this;
          co(function*() {
            const context = {}
            yield _this.refreshItem(shopBlacklist, historyToSave, historyPrices, subs, item, context);
            let steps = ['start', 'item_tojson', 'get_item_data', 'update_item_data', 'get_prom', 'get_price', 'calc_prom', 'create_history', 'calc_prom_score']
            let timeCost = {}
            let lastStep = null
            for (let step of steps) {
              if (!lastStep) {
                timeCost[step] = 0
              } else {
                timeCost[step] = context['t_' + step] - context['t_' + lastStep]
              }
              lastStep = step
            }
            console.log(Object.keys(timeCost).map(key => `${key}: ${timeCost[key]}`).join(', '))
          }).then(callback, callback)
        } catch (e) {
          console.error(e);
          itemsToRetry.push(item);
        }
      }, err => {
        done(err);
      })
    }
  }

  * refreshItems(items, tried) {
    tried = tried || 1;
    if (tried > 5) {
      yield this._alert.send(`保存价格记录失败`);
      return;
    }
    if (tried > 1) {
      logger.info('Retry:', tried);
    }
    let itemsToSave = [];
    let itemsToRetry = [];
    let itemsExpired = [];
    let feeds = [];
    let historyToSave = [];
    let historyPrices = yield this.getHistoryPrices(items);
    let subs = {};
    if (process.env.HX_TYPE === 'sub') {
      subs = yield this.getSubscriptions(items)
    }
    let shopBlacklist = yield this.getShopBlacklist();
    yield this.parallelRefreshItems(items, {
      shopBlacklist,
      historyToSave,
      historyPrices,
      itemsToRetry,
      subs
    });
    for (let item of items) {
      let changedAttributes = item.changedAttributes();
      if (changedAttributes) {
        itemsToSave.push(item);
      }
    }
    try {
      yield this.saveItems(itemsToSave);
    } catch (e) {
      itemsToRetry = itemsToRetry.concat(itemsToSave);
      logger.error('save item failed: ', e);
    }
    try {
      yield Bem.Object.saveAll(historyToSave);
    } catch (e) {
      logger.error('save price history failed: ', e);
    }
    // console.log(itemsToRetry);
    if (itemsToRetry.length > 0) {
      yield this.refreshItems(itemsToRetry, tried + 1);
    }
  }

  * getItemDataByJdId(itemId) {
    let cached = yield this.cache(`data-${itemId}`);
    if (cached) {
      return cached;
    }
    let url = `http://cdnware.m.jd.com/c1/skuDetail/apple/5.6.0/${itemId}.json`;
    let result;
    try {
      result = yield request({
        timeout: 3000,
        gzip: true,
        url
      });
      result = JSON.parse(result[1]);
    } catch (e) {
      logger.error(itemId, e);
      return null;
    }
    try {
      let wareInfo = result.wareInfo;
      let basicInfo = wareInfo.basicInfo;
      let shopInfo = wareInfo.shopInfo;
      let price = yield this.getItemPriceFromPage(itemId);
      let obj = {
        name: basicInfo.name,
        price: price,
        pic: basicInfo.wareImage[0].share,
        id: basicInfo.wareId,
        shop: shopInfo.shop ? shopInfo.shop.shopId + '' : '',
        vender: basicInfo.venderId,
        cat: basicInfo.category.split(';'),
        brand: basicInfo.brandID,
        platform: 'jd',
        self: !basicInfo.isPop,
        op: basicInfo.isOP
      };
      yield this.cache(`data-${itemId}`, obj);
      return obj;
    } catch (e) {
      logger.error(itemId, e);
      return null;
    }
  }

  * getItemData(item) {
    let itemId = item.get('jdId');
    return yield this.getItemDataByJdId(itemId);
  }

  * getProms(item, tried) {
    let itemId = item.get('jdId');
    tried = tried || 0;
    if (tried > this.MAX_TRY) {
      return null;
    }
    let itemData = yield this.getItemData(item);
    if (!itemData) {
      return null;
    }
    let url =
      `http://cd.jd.com/promotion/v2?skuId=${itemData.id}&shopId=${itemData.shop}&venderId=${itemData.vender}&cat=${itemData.cat.join(',')}&area=1_2800_4209_0`;
    let result;
    try {
      result = yield request({
        timeout: 3000,
        gzip: true,
        encoding: null,
        url
      })
    } catch (e) {
      logger.error(e);
      return yield this.getProms(itemId, tried + 1);
    }
    let body = result[1];
    try {
      body = iconv.decode(body, 'gbk');
      body = JSON.parse(body);
    } catch (e) {
      logger.error(e);
      return null;
    }
    let price = yield this.getItemPrice(item);
    // console.log(price)
    if (!price) {
      price = item.get('price');
    }
    let coupon = body.skuCoupon;
    let prom = body.prom.pickOneTag;
    let proms = [];
    let limit = (body.prom.tags || []).filter(tag => {
      return tag.name === '限购';
    });
    if (limit.length > 0) {
      limit = limit[0].content.match(/仅购买(\d+)件时享受优惠/);
      limit = (limit && limit[1]) || 9999;
    } else {
      limit = 9999;
    }
    if ((coupon && coupon.length > 0) || (prom && prom.length > 0)) {
      let couponDesc;
      if (coupon.length > 0) {
        couponDesc = coupon.map(item => item.allDesc || `满${item.quota}减${item.discount}`).join(',');
        let promItem = {
          name: '领券',
          desc: couponDesc,
          detail: coupon,
          t: new Date().toISOString()
        };
        let result = promCalc.calc(promItem.desc, price);
        //小于限购数量，才可用该优惠
        if (result.num <= limit) {
          if (result) {
            promItem.desc += '，买' + result.num + '件可享受优惠，总共需支付' + result.total + '元'
            promItem.price = result.price;
            promItem.buyCount = result.num;
            promItem.totalPay = result.total;
          }
          proms.push(promItem);
        }
      }
      if (prom.length > 0) {
        for (let item of prom) {
          let excludeNames = ['加价购', '满赠', '活动预告'];
          if (excludeNames.indexOf(item.name) !== -1) continue;
          let supportedProms = ['满减', '多买优惠', '跨店铺满减', '跨店铺满减进行中', '跨店铺满折进行中'];
          if (supportedProms.indexOf(item.name) !== -1) {
            let promItem = {
              name: item.name,
              desc: item.content,
              detail: item,
              t: new Date().toISOString()
            }
            let result = promCalc.calc(item.content, price);
            //如果结果没有返回price，说明这类促销需要被忽略
            if (result && result.price && result.num <= limit) {
              if (result) {
                promItem.desc = item.content + '，买' + result.num + '件可享受优惠，总共需支付' + result.total + '元'
                promItem.price = result.price;
                promItem.buyCount = result.num;
                promItem.totalPay = result.total;
                if (couponDesc) {
                  let addCounponResult = promCalc.calc(couponDesc, promItem.totalPay);
                  if (addCounponResult.num === 1) {
                    promItem.name = `${promItem.name} + 领券`;
                    promItem.totalPay = addCounponResult.total;
                    promItem.desc = item.content + `，领券再减${parseFloat(result.total - addCounponResult.total).toFixed(2)}，买` + result.num + '件可享受优惠，总共需支付' + addCounponResult.total + '元'
                    promItem.price = (parseFloat(promItem.totalPay) / promItem.buyCount).toFixed(2);
                  }
                }
              }
              proms.push(promItem);
            }
          }
        }
      }
    }
    return proms;
  }

  * getLinks(url, options) {
    let linkPatterns = [
      /\<a[^\>]*?href="\/\/[^\/]*?\.jd\.com[^\"]*"/g,
      /\<a[^\>]*?href="\/[^\/][^\"]*"/g
    ];
    let links = [];
    let body = yield this.getPageContent(url, options);
    if (!body) return links;
    linkPatterns.forEach(pattern => {
      let matched = body.match(pattern);
      if (matched) {
        matched = matched.map(item => {
          let link = item.match(/href="([^\"]*)"/)[1];
          if (link.indexOf('/') === 0) {
            link = URL.resolve(url, link);
          }
          return link;
        }).filter((link) => {
          let u = URL.parse(link);
          return this.blacklist.indexOf(u.hostname) === -1;
        });
        links = links.concat(matched);
      }
    });
    return links;
  }

  isItemLink(url) {
    let patterns = [
      /\/\/item\.jd\.com\/\d+\.html/,
      /\/\/item\.m\.jd\.com\/product\/\d+\.html/,
      /\/\/m\.jd\.com\/product\/\d+\.html/
    ]
    return patterns.some(pattern => {
      return !!url.match(pattern);
    })
  }

  getItemIdByURL(url) {
    let itemId;
    let patterns = [
      /\/\/item\.jd\.com\/(\d+)\.html/,
      /\/\/item\.m\.jd\.com\/product\/(\d+)\.html/,
      /\/\/m\.jd\.com\/product\/(\d+)\.html/,
      /\/\/item\.jd\.hk\/(\d+).html/,
      /\/\/mitem\.jd\.hk\/product\/(\d+).html/,
      /\/\/wqitem\.jd\.com\/item\/view\?sku=(\d+)/,
      /\/\/re\.m\.jd\.com\/cps\/item\/(\d+).html/,
      /\/\/re\.jd\.com\/cps\/item\/(\d+).html/,
      /\/\/wq\.jd\.com\/item\/view\?sku=(\d+)/
    ]
    patterns.some(pattern => {
      let matched = url.match(pattern);
      if (matched) {
        itemId = matched[1];
        return true;
      }
      return false;
    })
    return itemId;
  }

}

module.exports = CrawlJd;
