'use strict'

var assert = require("assert");
var Crawl = require('./');
let co = require('co');
let thunkify = require('thunkify');
const Bem = require('bemcloud-storage');
let crawl = new Crawl({
  Bem
});

// Bem._config.APIServerURL = 'http://localhost:4001';
Bem._config.APIServerURL = 'http://wxapi.bemcloud.com';

Bem.init({
  appId: 'dmVkYGPFKwMuCEZsU2mwEmpTkprNVQdM',
  javascriptKey: 'G3u5FtVYMqcBXrvHABdKXWa6unY6LJFn',
  masterKey: 'nF5n634E6ZWwUErhzm2EWCpiaQjHmEpn'
});

describe('crawl', function() {

  this.timeout(10 * 1000);

  describe('getProms', function() {
    it('should return proms info array', function(done) {
      co(function*() {
        let itemId = crawl.getItemIdByURL('http://item.jd.com/10828544545.html')
        let query = new Bem.Query('Items');
        let item = (yield query.equalTo('jdId', itemId).find())[0];
        let proms = yield crawl.getProms(item);
        assert.notEqual(proms.length, 0);
      }).then(done, done);
    });
  });

  describe('getLinks', function() {
    it('should return link array', function(done) {
      co(function*() {
        let result = yield crawl.getLinks('http://www.jd.com');
        assert.notEqual(result.length, 0);
      }).then(done, done);
    });
  });

  describe('isItemLink', function() {

    it('should return true checking pc url', function(done) {
      let result = crawl.isItemLink('http://item.jd.com/832264.html');
      assert.equal(result, true);
      done();
    });

    it('should return true checking h5 url', function(done) {
      let result = crawl.isItemLink('http://item.m.jd.com/product/832264.html');
      assert.equal(result, true);
      done();
    });

    it('should return true checking app url', function(done) {
      let result = crawl.isItemLink(
        'http://m.jd.com/product/10074808892.html?resourceType=jdapp_share&resourceValue=CopyURL&utm_source=iosapp&utm_medium=appshare&utm_campaign=t_335139774&utm_term=CopyURL');
      assert.equal(result, true);
      done();
    });

  });

  describe('getItemData', function() {
    // it('should return item data', function(done) {
    //   co(function*() {
    //     let itemId = crawl.getItemIdByURL('http://item.jd.com/832264.html')
    //     let result = yield crawl.getItemData(itemId);
    //     assert.equal(result.name, '【京东超市】金龙鱼 粳米 珍珠米 优质东北大米 5kg');
    //   }).then(done, done);
    // });

    it('should return item data by item id', function(done) {
      co(function*() {
        let itemId = crawl.getItemIdByURL('http://item.jd.com/832264.html')
        let result = yield crawl.getItemDataByJdId(itemId);
        assert.equal(result.name, '【京东超市】金龙鱼 粳米 珍珠米 优质东北大米 5kg');
      }).then(done, done);
    });
  });

  describe('getItemIdByURL', function() {

    it('should return item id using pc url', function(done) {
      co(function*() {
        let result = crawl.getItemIdByURL('http://item.jd.com/832264.html');
        assert.equal(result, '832264');
      }).then(done, done);
    });

    it('should return item id using h5 url', function(done) {
      let result = crawl.getItemIdByURL('http://item.m.jd.com/product/832264.html');
      assert.equal(result, '832264');
      done();
    });

    it('should return item id using app url', function(done) {
      let result = crawl.getItemIdByURL(
        'http://m.jd.com/product/832264.html?resourceType=jdapp_share&resourceValue=CopyURL&utm_source=iosapp&utm_medium=appshare&utm_campaign=t_335139774&utm_term=CopyURL');
      assert.equal(result, '832264');
      done();
    });

    it('should return item id using wq url', function(done) {
      let result = crawl.getItemIdByURL(
        'https://wq.jd.com/item/view?sku=1297484&mktid=1273&shopid=2596911&enc=qxapHl-Sn5Ayi1aotUDFjDNnXdZUy7wRrALLMjboD-JDnV2yV_F6s3FjSwkvyHhib1JcW8FlXB_EocQqn0Al1m4rJP27mgTGDP_qqMyj29NT_86WiSM0JA==&tk=19ac33fb5870c8d5c699ed7974d7df0d&share=0&fs=1&gt=1490940219&v=1&cu=true&utm_source=kong&utm_medium=unionjingxiangjie&utm_campaign=t_2010472270_&utm_term=b210ce2b6a8b45bcb7d8106b06ee6f07&abt=3#main'
      );
      assert.equal(result, '1297484');
      done();
    });

  });

  describe('getItemPrice', function() {
    it('should return item price', function(done) {
      co(function*() {
        let itemId = crawl.getItemIdByURL('http://item.jd.com/1954210.html');
        let result = yield crawl.getItemPrice(itemId);
        assert.equal(result, 99.5);
      }).then(done, done);
    });
  });

  describe('refresh items', function() {
    this.timeout(30000);
    it.only('should refresh items', function(done) {
      co(function*() {
        let itemIds = ['4682259',];
        let query = new Bem.Query('Items');
        let items = yield query.containedIn('jdId', itemIds).find();
        yield crawl.refreshItems(items);
        // assert.equal(result, '30.90');
      }).then(done, done);
    });
  });

});
